import psyco, time, sys, PyLucene, os

import Modules.BerkeleyDBXML as BerkeleyDBXML
import Modules.Logger as Logger
import Modules.Indexer as Indexer
import cElementTree as ElementTree
from threading import Thread, Event
from Modules.CommonFunctions import *

class Main:
    def __init__(self): #init application
        self.version = "Farscape XML Indexer Prototype v0.90\nNTNU - Masterproject 2006\nTrond A Myklebust\n"
        self.xmlpathSmall = "xml/" #directory for XML data
        self.logger = Logger.Logger() #logger
        self.bdxml = BerkeleyDBXML.BerkeleyDBXML() #berkeley db xml class
        self.indexer = Indexer.Indexer() #Indexer class
        self.softQuit = Event()
        self.QueueEvent = Event()
        self.PauseEvent = Event()
        self.bdxmlpath = "bdbxml/xml" #berkeley db xml index file
        self.start()
 
    def start(self): #Start threads and print version data, threads are PyLucene.PythonThread as required by PyLucene for memory handling
        self.logger.clearScreen()
        print self.version
        self.processThread = PyLucene.PythonThread(target=self.processThread)
        self.processThread.start()
        self.controllerThread = PyLucene.PythonThread(target=self.controllerThread)    
        self.controllerThread.start()
        self.commandThread = PyLucene.PythonThread(target=self.commandThread)  
        self.commandThread.start()
       
    #Starts/stops parsing of documents from berkeley dbxml if too many documents has been parsed and not added to PyLucene
    def controllerThread(self):
        self.logger.Output("controllerThread", "controllerThread started")
        min = 10000 #lower threshold
        max = 20000 #upper threshold
        while not self.softQuit.isSet(): #quit if event is set
            time.sleep(1)
            if len(self.indexer.queue) > max and not self.PauseEvent.isSet(): #pause parsing from berkeley if max threshold is reached
                self.PauseEvent.set()
            elif len(self.indexer.queue) < min and self.PauseEvent.isSet(): #start parsing from berkeley db xml again
                self.PauseEvent.clear() 
        self.logger.Output("controllerThread", "controllerThread stopped")
      
    #Thread for acting to commands based on menuselection.
    def commandThread(self):
        self.logger.Output("commandThread", "commandThread started")
        self.run = True
        
        while self.run:
            response = self.commandMenu()
            if response == "1": #add documents to Berkeley DBXML
                self.bdxml.run(self.bdxmlpath, self.logger.Output, self.xmlpathSmall)
            elif response == "2": #start indexing
                self.indexer.start(self.QueueEvent, self.PauseEvent, self.logger) #(self.QueueEvent, self.PauseEvent, self.logger, self.bdxmlpath, self.indexDB, self.pathindexDB, self.pathindexRDB, self.docTreeDB, self.filenameDB, self.datatypeDB)
            elif response == "3":
                deleteFiles(os.getcwd() + "\\bdb")  #directory for berkeley db data
                deleteFiles(os.getcwd() + "/bdbxml") #directory for Berkeley dbxml index
                deleteFiles(os.getcwd() + "/LuceneIndex") #directory for LuceneIndex
            elif response == "4": #Show log
                self.logger.showLog()
            elif response == "5": #toggle verbose
                self.logger.toggleVerbose()
            elif response == "6": #quit application
                self.quit()
            else:
                print "-Command not valid"
        self.logger.Output("commandThread", "commandThread stopped")
        
    #Thread thats while app is still running checks the document queue for new items and calls a function that adds them to PyLucene
    def processThread(self):
        self.logger.Output("processThread", "processThread started")           
        while not self.softQuit.isSet():
            self.QueueEvent.wait( 5 )
            while len(self.indexer.queue) > 0: #While something in queue, index content
                (docName, pathID, content, path, treeLoc, datatype, complextype, attrib, annotation) = self.indexer.queue[0]
                del self.indexer.queue[0]
                self.indexer.addToIndex(docName, pathID, content, self.indexer.indexDB, path, treeLoc, datatype, complextype, attrib, annotation)
            self.QueueEvent.clear() 
        self.logger.Output("processThread", "processThread stopped")

    #Prints menu to screen
    def commandMenu(self):
        print "\nMain Menu:"
        print "-------------------------------------"
        print "1. Add collection to Berkeley DBXML"
        print "2. Start indexing"
        print "3. Clear indexes/databases"
        print "4. Show log"
        print "5. Verbose on/off"
        print "6. Quit program"
        print "-------------------------------------"
        return raw_input("> ")
        
    #Stop threads and quit app
    def quit(self):
        self.softQuit.set()
        self.run = False
                 
#Invoke psyco optimizer
psyco.full()     
#Start application
main = Main()