import os, PyLucene, zlib
import Lucene.myAnalyzer as myAnalyzer
import Lucene.ScoreMethod as ScoreMethod
import cElementTree as ElementTree

from bsddb import *
from dbxml import *
from PyLucene import IndexSearcher, FSDirectory

LuceneDir = 'LuceneIndex'
#bdxmlpath = os.getcwd().replace("\\", "/") + "/" + "bdbxml/xml" + ".dbxml"
bdxmlpath = "bdbxml/xml" + ".dbxml"
pathindexRname = "bdb/DocPathRIndex.db"
pathindexname = "bdb/DocPathIndex.db"
doctreename = "bdb/DocTreeIndex.db"
docfilenamename = "bdb/docFileName.db"
datatypename = "bdb/datatypeName.db"

class Database:
    def __init__(self):
        self.pathindexDB = None
        self.pathindexRDB = None
        self.docTreeDB = None
        self.docFileNameDB = None
        self.datatypeDB = None
        self.indexDB = None
        self.container = None

    #Open indexes for read only, required for writing to index while indexing
    def OpenIndex(self):
        analyzer = myAnalyzer.myAnalyzer()

	#Berkeley db xml index
        mgr = XmlManager(DBXML_ALLOW_EXTERNAL_ACCESS)
        uc = mgr.createUpdateContext()
        container = mgr.openContainer(bdxmlpath)
        qc = mgr.createQueryContext()
        qc.setEvaluationType(XmlQueryContext.Lazy) #lazy, get data when requested
        results = mgr.query("collection('" + bdxmlpath + "')", qc) #retrieve all documents
        
        if not os.path.exists(LuceneDir): #if PyLucene directory doesnt exist, create
            os.mkdir(LuceneDir)
        
        #PyLucene index
        store = PyLucene.FSDirectory.getDirectory(LuceneDir, True)
        indexDB = PyLucene.IndexWriter(store, analyzer, True)
        indexDB.minMergeDocs = 1000
        indexDB.mergeFactor = 500
        
        #Document tree data
        docTreeDB = db.DB()
        docTreeDB.set_cachesize(0, 20 * 1024 * 1024)
        docTreeDB.open(doctreename, None, db.DB_RECNO, db.DB_CREATE)

	#Document Filename - Berkeley DB XML ID index
        docFileNameDB = db.DB()
        docFileNameDB.set_cachesize(0, 20 * 1024 * 1024)
        docFileNameDB.open(docfilenamename, None, db.DB_RECNO, db.DB_CREATE)

	#XPath index ID -> Path
        pathindexDB = db.DB()
        pathindexDB.set_cachesize(0, 20 * 1024 * 1024)
        pathindexDB.open(pathindexname, None, db.DB_RECNO, db.DB_CREATE)
	
	#XPath index Path -> ID
        pathindexRDB = db.DB()
        pathindexRDB.set_cachesize(0, 20 * 1024 * 1024)
        pathindexRDB.open(pathindexRname, None, db.DB_BTREE, db.DB_CREATE)

        #Index of datatypes
        datatypeDB = db.DB()
        datatypeDB.set_cachesize(0, 20 * 1024 * 1024)
        datatypeDB.open(datatypename, None, db.DB_BTREE, db.DB_CREATE)
        
        self.indexDB = indexDB
        self.pathindexDB = pathindexDB
        self.pathindexRDB = pathindexRDB
        self.docTreeDB = docTreeDB
        self.docFileNameDB = docFileNameDB
        self.datatypeDB = datatypeDB
        self.container = container

        return (bdxmlpath, self.indexDB, self.pathindexDB, self.pathindexRDB, self.docTreeDB, self.docFileNameDB, self.datatypeDB, results)

    #Open indexes for read only, required for searching in index
    def OpenSearch(self):

        print "-Accessing Berkeley DBXML index"
        mgr = XmlManager(DBXML_ALLOW_EXTERNAL_ACCESS)
        container = mgr.openContainer(bdxmlpath)
        qc = mgr.createQueryContext();
        
        print "-Accessing PyLucene index"
        analyzer = myAnalyzer.myAnalyzer()
        directory = FSDirectory.getDirectory(LuceneDir, False)
        searcher = IndexSearcher(directory)
        searcher.setSimilarity(ScoreMethod.SimpleSimilarity())
        reader = PyLucene.IndexReader.open(directory)

        pathindexDB = db.DB()
        pathindexname = "bdb/DocPathIndex.db"
        pathindexDB.set_cachesize(0, 20 * 1024 * 1024)
        pathindexDB.open(pathindexname, None, db.DB_RECNO, db.DB_CREATE)

        print "-Accessing docFileNameindex"
        docFileName = db.DB()
        docFileName.set_cachesize(0, 20 * 1024 * 1024)
        docFileName.open("bdb/docFileName.db", None, db.DB_RECNO, db.DB_CREATE)
        docsFileName = []
        for id in docFileName.keys():
            docsFileName.append(docFileName[id])
        docFileName.close()

        print "-Accessing DocTree index"
        docTreeDB = db.DB()
        docTreeDBname = "bdb/DocTreeIndex.db"
        docTreeDB.set_cachesize(0, 20 * 1024 * 1024)
        docTreeDB.open(docTreeDBname, None, db.DB_RECNO, db.DB_CREATE)
        docsStore = {}
        for doc in docTreeDB.keys():
            docsStore[doc] = ElementTree.XML(zlib.decompress(docTreeDB[doc]))
        docTreeDB.close()
    
        print "-Accessing datatype index"
        datatypeDB = db.DB()
        datatypeDB.set_cachesize(0, 20 * 1024 * 1024)
        datatypeDB.open(datatypename, None, db.DB_BTREE, db.DB_CREATE)
        self.datatypeDB = datatypeDB
        
        self.pathindexDB = pathindexDB
        self.reader = reader
        self.searcher = searcher
        
        return (bdxmlpath, analyzer, container, mgr, searcher, reader, pathindexDB, docsFileName, docsStore, qc, datatypeDB)

    #Close open indexes
    def CloseSearch(self):
        self.pathindexDB.close()
        self.reader.close()
        self.searcher.close()
        self.datatypeDB.close()
          
    #Close open indexes, important. If added new documents to index they will not be saved if not closed properly
    def CloseIndex(self):
        if self.indexDB:
            self.indexDB.close()
        if self.pathindexDB:
            self.container.close()
            self.pathindexDB.close()
            self.pathindexRDB.close()
            self.docTreeDB.close()
            self.docFileNameDB.close()
            self.datatypeDB.close()