import datetime, time, sys, sre, PyLucene, re
import cElementTree as ElementTree
import Modules.Database as DataBase
import Modules.DataTypeNormalizer as DataType
import Annotations as Annotations

from CommonFunctions import *
from XSV import *
from dbxml import *
from threading import Thread, Event, RLock

#timer for optimizing PyLucene index
class Ticker(object):
    def __init__(self):
        self.tick = True

    def run(self):
        while self.tick:
            sys.stdout.write('.')
            sys.stdout.flush()
            time.sleep(1.0)
        sys.stdout.write('\n')
        
class Indexer:
    def __init__(self): #initiate databases and events
        self.datatype = DataType.DataType()
        self.database = DataBase.Database()
        self.indexDB = None
        self.logger = None
        self.queue = []
        self.QueueEvent = None
        self.PauseEvent = None
        self.xmlschemalocation = "{http://www.w3.org/2001/XMLSchema-instance}schemaLocation"

    def start(self, QueueEvent, PauseEvent, logger): #start indexing, finish with optimizing index
        (self.xmldbpath, self.indexDB, self.pathindexDB, self.pathindexRDB, self.docTreeDB, self.filenameDB, self.datatypeDB, results) = self.database.OpenIndex()
        self.logger = logger
        self.QueueEvent = QueueEvent
        self.PauseEvent = PauseEvent

        self.logger.Output("startIndexer", "Indexing Berkeley DB XML collection")
        print "-Indexing documents in Berkeley DBXML database"
        start = datetime.datetime.now() 
        results.reset()
        counter = self.indexDocs(results, self.pathindexDB, self.pathindexRDB, self.docTreeDB, self.filenameDB, self.datatypeDB)                
        while len(self.queue) > 0:
            time.sleep(1)
        self.optimizeIndex() #run optimizer
        print "\n-Finished indexing documents in Berkeley DBXML database"
        self.logger.Output("Indexer", "Time used: %s" % (datetime.datetime.now() - start))
        self.logger.Output("Indexer", "Done indexing")
        self.close()

    def close(self): #close indexes when finished indexing
        self.database.CloseIndex()
        
    def indexDocs(self, results, pathindexDB, pathindexRDB, docTreeDB, filenameDB, datatypeDB):
        docCounter = 0
        pathCounter = 0

        for value in results: #for each document in Berkeley DB XML
            docName = value.asDocument().getName() #get name
            self.logger.Output("IndexDocs", "Indexing: " + docName)
            (docDOM, PSVI, value) = self.getDocument(value)    #get document content, parse XML Schema (PSVI), docdom contains DOM model
            if docDOM:
                rootTag = docDOM.tag.split("}")
                path = rootTag[-1] + "[1]"
                cpath = serialize_via_marshal(u"/" + path)
                if PSVI and PSVI[1]: #if PSVI exists, check if there is any tags that should not be indexes, if so, remove and reparse documentcontent
                    anno = Annotations.Annotations(PSVI[1][u"/" + path][5])
                    RemoveTags = anno.getRemoveTags()
                    for tag in RemoveTags: #remove tags, keep content
                        value = value.replace("<%s>" % tag, "")
                        value = value.replace("</%s>" % tag, "")
                    if len(RemoveTags) > 0: #reparse documentcontent
                        (docDOM, PSVI, value) = self.getDocument(value)
                if not pathindexRDB.has_key(cpath): #store xpath -> id and id ->xpath
                    pathCounter += 1
                    pathindexDB.put(pathCounter, cpath)
                    pathindexRDB.put(cpath, "%s" % pathCounter)
                    docTree = ElementTree.Element("r%s" % pathCounter) #generate document tree using Effbot elementtree
                else:
                    docTree = ElementTree.Element("r%s" % pathindexRDB.get(cpath))
                pathCounter = self.walk(PSVI[1], docDOM, path, "%s" % docCounter, pathindexDB, pathindexRDB, pathCounter, self.queue, docTree, "") #walk XML tree of a document
                docTreeDB[docCounter + 1] = zlib.compress(ElementTree.tostring(docTree, "UTF-8"))#serialize_via_pickle(docTree)
                filenameDB[docCounter + 1] = docName #store document id -> documentname
                del docDOM
                docCounter += 1
        return docCounter
        
    def getDocument(self, value): #parse XML content
        if not type(value) in [unicode, str]:
            value = value.asString()
        try:
            docDOM = ElementTree.XML(value)
        except Exception,e:
            value = """<?xml version="1.0" ?>""" + value[value.find("<books"):] #INEX
            docDOM = ElementTree.XML(value)
        PSVI = commandLine.parseAndRun([value])
        if PSVI[2] != 0:
            print "Could not validate the whole document, parts of the document may not be indexed"         
        return (docDOM, PSVI, value)
                
    def optimizeIndex(self): #optimize PyLucene index
        ticker = Ticker()
        self.logger.Output("optimizeIndex", "Optimizing index")
        Thread(target=ticker.run).start()
        self.indexDB.optimize()
        ticker.tick = False
        self.logger.Output("optimizeIndex", "Finished optimizing")


    def walk(self, PSVI, parent, path, docName, pathindexDB, pathindexRDB, pathCounter, queue, docTree, treeLoc): #walk XML tree, recursive
        counter = 0
        for node in parent: #for each childnode
            treeLoc += " %s" % counter
            path = self.make_xpath_expr(parent, path, node)
            cpath = serialize_via_marshal("/" + path)
            pPSVI = (PSVI and PSVI[u"/" + path] or ["","","","","", None])
            annotation = Annotations.Annotations(pPSVI[5])  #get annotations from XML schema
            content = (node.text and node.text or "")
            concenate = annotation.getMixedConcenate()
            for child in node:
                if concenate and child.tag in concenate: #concenate content if given in xs:annotation
                    content += child.text + " "
                if pPSVI[3] == "mixed":
                    content += (child.tail and child.tail or "")
            id = 0
            if not pathindexRDB.has_key(cpath): #store xpath -> id, id->xpath for node
                pathCounter += 1
                id = pathCounter
                pathindexDB.put(pathCounter, cpath)  
                pathindexRDB.put(cpath, "%s" % pathCounter) 
            else:
                id = int(pathindexRDB.get(cpath))
            queue.append((docName, id, content, path, treeLoc, pPSVI[3], pPSVI[4], "", annotation)) #append xpath, content and PSVI to queue for indexing by PyLucene
            tree = ElementTree.SubElement(docTree, "r%s" % id)
            
            if len(node.attrib.keys()) > 0: #if attrib, add to queue for indexing by PyLucene
                for key in node.attrib.keys(): 
                    attpath = path + "/@" + key
                    attcpath = serialize_via_marshal("/" + attpath)
                    id = 0
                    if not pathindexRDB.has_key(attcpath):
                        pathCounter += 1
                        id = pathCounter
                        pathindexDB.put(pathCounter, attcpath)  
                        pathindexRDB.put(attcpath, "%s" % pathCounter) 
                    else:
                        id = int(pathindexRDB.get(attcpath))
                    aPSVI = (PSVI and PSVI.get(u"/" + attpath, ["","","","","", None]) or ["","","","","", None])
                    queue.append((docName, id, node.attrib[key], attpath, treeLoc, aPSVI[3], aPSVI[4], key, annotation))
            self.QueueEvent.set()
            pathCounter = self.walk(PSVI, node, path, docName, pathindexDB, pathindexRDB, pathCounter, queue, tree, treeLoc) #check if childelement got its own childs
            counter += 1
            treeLoc = treeLoc[:treeLoc.rfind(" ")]
            path = path[0:path.rindex("/")]
            while self.PauseEvent.isSet(): #if pause is set, wait until nodes has been indexed by PyLucene
                time.sleep(1)
        return pathCounter
        
    def make_xpath_expr(self, parent, path, node): #create xpath expression for node
        name = node.tag.split("}")[-1]
        no = 0
        count = 0
        for testNode in parent:
            tagsplit = testNode.tag.split("}")
            if testNode.tag.split("}")[-1] == name:
                count += 1 
                if testNode == node:
                    no = count
                    break
        return path + "/%s[%d]" % (name, no)

    #index node using PyLucene
    def addToIndex(self, docName, pathID, content, indexDB, ancestors, treeLoc, datatype, complextype, attrib, annotation):
        content = content.replace("\n", " ").strip()
        if not content:
            return
        doc = PyLucene.Document() #create new "document" in PyLucene
        ancestors = ancestors.replace("[", "<")
        ancestors = ancestors.replace("]", ">")
        ancestors = sre_html.sub('', ancestors)
        tagName = ancestors.split("/")[-1]
        ancestors = ancestors[:-len(tagName)]
        doc.add(PyLucene.Field("treeLoc", treeLoc.strip(), True, False, False)) #add location in tree to index
        doc.add(PyLucene.Field("docID", docName, True, False, False)) #add document id to index
        doc.add(PyLucene.Field("pathID", "%s" % pathID, True, False, False)) #add pathid to index
        doc.add(PyLucene.Field("content", content, False, True, True)) #add content with case to index
        doc.add(PyLucene.Field("contentL", content, False, True, True))   #add content without case to index
                    
        if not datatype or datatype in ["mixed", "elementOnly"]: #if datatype given in xs:annotation and not using simpletype, get data
          datatype = annotation.getDataType()
        if datatype: #store content in PyLucene with datatype as field
            datatype = datatype.encode('ascii')
            content = self.datatype.Modify(content, datatype, annotation.getFormat())
            doc.add(PyLucene.Field(datatype, content, False, True, True))
        else:
            self.logger.Output("addToIndex", (int(docName), tagName, content, datatype, complextype, annotation))
        indexDB.addDocument(doc) #add document to PyLucene 
        del doc

